﻿module MasterDataManagement

open System
open System.Threading
open System.Threading.Tasks
open System.Collections
open System.Collections.Generic
open System.Text

/// Name parsing
let ParseName (strName : string) = 
    let splitNames = strName.Trim().Split([|" "|], StringSplitOptions.None)
    match splitNames.[0] with
    | "Mr" | "Mr." | "Mrs." | "Mrs" | "Dr." | "Dr" | "Kum." | "Kum" -> splitNames.[1]
    | _ -> splitNames.[0]

// Exhaustive lookup for street names, state names, zip codes etc are available from companies which compile and sell such data.
// Companies that make customized Master Data Management solutions buy such country specific data for street names, city names, state names etc
// Here I present just a subset of that in order to explain the algorithm 

//We assume that the table for zip codes is obtained from database and stored in a list.
//Right now for example sake I am adding a few sample zip codes.
let zipCodes = ["12098"; "13456"; "23458"; "78906"; "89067"; "98765"; "700107"]

//We assume that the table for apartment types are  obtained from database and stored in a list.
//Right now for example sake I am adding a few sample apartment types.
let apartmentTypes = ["Suite"; "Single"; "Double"; "Presidential"; "Kitchen"; "Modular"]

//We assume that the table for city names are  obtained from database and stored in a list.
//Right now for example sake I am adding a few sample city names.
let cities = [ "Kolkata"; "New York"; "Los Angeles"; "Austin"; "New Jersey" ]

//We assume that country specific table for states is obtained from database and added to array as tuples.
//Right now for the sake of explanation, I am adding a few state names.
let stateNameTable = [ ("AZ", "Arizona"); ("FL", "Florida"); ("CA", "California"); ("WB", "West Bengal") ]

//Searching for valid city name, state name or zip code 
let MatchCityStateZip (addressPart : string) = 
    //Match with a state
    let stateMatchFound = stateNameTable |> List.exists (fun (part1,part2) -> part1 = addressPart || part2 = addressPart) 
    //Match with a city
    let cityMatchFound = cities |> List.exists (fun city -> city = addressPart)
    //Match with a zip code
    let zipMatchFound = zipCodes |> List.exists (fun zipCode -> zipCode = addressPart)
    stateMatchFound || cityMatchFound || zipMatchFound

///The main address parsing algorithm.
///The parameter strAddress is the concatenation of all address lines
let ParseAddress (strAddress : string) = 
    let mutable finalAddress = strAddress
    //Split the address
    let addressParts = strAddress.Split([|","|], StringSplitOptions.None)
    //Remove city, state and zip information from the address
    for i = 0 to addressParts.Length - 1 do
        let currPart = addressParts.[i].Trim()
        if MatchCityStateZip currPart then
            //Index of current address part in original string
            let currIndex = finalAddress.IndexOf currPart
            //Remove city, state, zip information alongwith the following whitespace or comma
            finalAddress <- finalAddress.Remove(currIndex, currPart.Length)
    let finalAddress = finalAddress.Replace(", ,",",")
    let finalAddress = finalAddress.TrimEnd([|','; ' '|])
    finalAddress


//Absolute Match Routine
let AbsoluteMatch (attr1:string, attr2:string) = 
    let attrRec01 = attr1.Trim().ToLower()
    let attrRec02 = attr2.Trim().ToLower()
    match attrRec01, attrRec02 with
    | "", "" -> 100
    | _, _ -> if attrRec01 = attrRec02 then 100 else 0

//Partial Match Routine
let PartialMatch (attr1:string, attr2:string) = 
    let attrRec01 = attr1.Trim().ToLower()
    let attrRec02 = attr2.Trim().ToLower()
    match attrRec01, attrRec02 with
    | "", "" -> 65
    | _, "" | "", _ -> 75
    | _, _ -> if attrRec01 = attrRec02 then 100 else 0

//Soundex Match
let SoundexMatch (attr1:string, attr2:string) = 
    let conv c = 
        match c with
        | 'A' | 'E' | 'I' | 'O' | 'U' | 'W' | 'Y' -> ' '
        | 'B' | 'F' | 'P' | 'V' -> '1'
        | 'C' | 'G' | 'J' | 'K' | 'Q' | 'S' | 'X' | 'Z' -> '2'
        | 'D' | 'T' -> '3'
        | 'L' -> '4'
        | 'M' | 'N' -> '5'
        | 'R' -> '6'
        | _ -> c

    let convertSoundex (inp:string) = 
        // Capitalize all letters in the string
        let chars = inp.ToUpper().ToCharArray() 
        let chars = 
            [| // Retain the first letter of the string
               yield chars.[0] 
               // Keep the first character, and remove pairwise-duplicates
               // Change letters from the predetermined sets to corresponding number 
               for (c1,c2) in Seq.pairwise (Seq.map conv chars) do
                   // Remove all consecutive pairs of duplicate digits and blanks from the string 
                   if c1 <> c2 && c2 <> ' ' then yield c2 |]
        // Convert back to a string
        String chars

    //Retain first four characters of resultant strings padded with trailing zeros if needed; leave unchanged if any string is blank
    let adjustResult (result:string) = 
        match result.Length with 
        | n when n >= 4 -> result.Substring(0, 4)
        | 0 -> result
        | n -> result + String.replicate (4 - n) "0"

    let attr1Result = attr1 |> convertSoundex |> adjustResult
    let attr2Result = attr2 |> convertSoundex |> adjustResult

    match attr1Result, attr2Result with
    | "", "" -> 0
    | "", _ | _, "" -> 50
    | _, _ -> if (attr1Result = attr2Result) then 100 else 0

let WeightedAverage results = 
    let cumulativeWeight = results |> Array.sumBy (fun (r, weight) -> r * weight) 
    let totalWeight      = results |> Array.sumBy (fun (r, weight) -> weight)     
    cumulativeWeight / totalWeight

/// Aggregate match score
/// Calling the Match Routing which in turn calls Absolute Match, Partial Match and Soundex Match routines in parallel
let FindAggregateMatchScore row = 
    let resultsWithWeights = 
        Async.Parallel [ async { return AbsoluteMatch row, 5 } 
                         async { return PartialMatch row, 5 } 
                         async { return SoundexMatch row, 4} ]
        |> Async.RunSynchronously

    WeightedAverage resultsWithWeights 

/// Percentage match score
let FindPercentageMatchScore(rows : seq<string * string>) = 
    let results = 
        Async.Parallel [ for row in rows -> async { return FindAggregateMatchScore row } ] 
        |> Async.RunSynchronously

    //Apply a weight of 5 for the first attribute and a weight of 4 for second and a weight of 3 for all other attributes
    let resultsWithWeights = results |> Array.mapi (fun i r -> r, (match i with 0 -> 5 | 1 -> 4 | _ -> 3)) 

    WeightedAverage resultsWithWeights 

//We assume that the table for street names is obtained from database and stored in a list
//Right now for example sake I am adding a few sample street names to a list
let streetNames = ["East Drachman"; "Rajdanga Main Road"; "Rashbehari Avenue"; "Goldpark"; "Pavilion boulevard"; "Riata Trace Parkway"]

///Lookup Match
let LookupMatch (streetName : string) = 
    if streetNames |> List.exists (fun s -> s = streetName)  then 
        100 
    else 
        0

[<EntryPoint>]
let main (args : string[]) = 
    // Ensure we get sufficient CPU parallelism by default
    System.Threading.ThreadPool.SetMinThreads(10,10) |> ignore

    let resName = ParseName "Mrs. John"
    printfn "Name is %s" resName

    //Testing address parsing
    let resAddress = ParseAddress @"DF-I,B-2/4,PURBA ABASAN, 1582/1, Rajdanga Main Road, Kolkata, 700107, West Bengal"
    printfn "Final Address : %s" resAddress

    let sampleRow = (" owyiou", "Pintowc")
    
    //Testing absolute matching
    let abs = AbsoluteMatch sampleRow
    printfn "Absolute Match Score %d" abs

    //Testing partial matching
    let par = PartialMatch sampleRow
    printfn "Partial Match Score %d" par

    //Testing soundex matching
    let sou = SoundexMatch sampleRow
    printfn "Soundex match: %d" sou

    //Testing aggregate match score
    let agg = FindAggregateMatchScore sampleRow
    printfn "Aggregate Score :%d" agg

    //Testing percentage match score
    let rows = [ ("Ambar", "ambar"); ("ambar", "ambar"); ("Pinto", "Pintow")]
    let results = FindPercentageMatchScore rows
    printfn "Results %d" results

    //Testing lookup match score
    let street = LookupMatch "Rajdanga Mai Road"
    printfn "Lookup Match: %d" street

    Console.ReadKey true |> ignore
    0

